indexing

   description: "Class that can parse a mime encoded (multipart) message."

   author: "Berend de Boer"
   date: "$Date: 2001/12/27 $"
   revision: "$Revision: #1 $"

   known_bugs:
   "1. Header lines are limited to 1024 bytes max."

class

   EPX_MIME_PART


inherit

   STDC_BASE

   STDC_FILE_SYSTEM

   STDC_CURRENT_PROCESS

creation

   make,
   make_from_header


feature -- creation

   make (a_reader: EPX_READER) is
      require
         valid_reader: a_reader /= Void
         -- reader is positioned at first character MIME header
      do
         create fields.make (0, -1)
         reader := a_reader
         create last_line.allocate (128)
      end

   make_from_header (a_reader: EPX_READER; a_mime_header: STRING) is
         -- a single line mime header is given in `a_mime_header'
         -- after that, mime message is read as usual
      require
         valid_header: a_mime_header /= Void and then not a_mime_header.is_empty
      do
         make (a_reader)
         header := a_mime_header
      end


feature -- state

   body: EPX_MIME_BODY

   save_body: BOOLEAN
         -- Option to save mime body in a file, if it is a file.
         -- A body is a file when it has a filename parameter
         -- body is saved to a temporary file and the mime part gets
         -- an extra parameter temporary_filename

   fields: ARRAY[EPX_MIME_FIELD]


feature -- state change

   set_save_body (a_value: BOOLEAN) is
         -- write mime bodies to a file
      do
         save_body := a_value
      end


feature {NONE}  -- state used during parsing

   boundary: STRING
         -- boundary to use when multipart string

   header: STRING
         -- header to use when header not present in input

   last_line: EPX_SMART_BUFFER
         -- we store last read line here when reading multipart bodies
         -- as we potentially store binary data, a STRING will not do

feature -- parsing

   parse is
         -- start parsing mime message
      do
         if header = Void then
            read_mime_header
            -- `input' is set to first line after the first blank line
            -- encountered
            check
               input_cursor_ok: reader.last_string.is_empty
            end
         else
            decode_header_field (header)
         end
         read_mime_body (boundary)
      end

   set_boundary (a_boundary: STRING) is
      do
         boundary := a_boundary
      end


feature -- queries

   field (a_name: STRING): EPX_MIME_FIELD is
         -- return field if present in MIME header
      local
         lname: STRING
         i: INTEGER
         found: BOOLEAN
      do
         from
            create lname.make_from_string (a_name)
            lname.to_lower
            found := False
            i := fields.lower
         until
            found or else i > fields.upper
         loop
            Result := fields.item (i)
            found := Result.name.is_equal (lname)
            i := i + 1
         end

         check
            found implies Result /= Void
         end

         if not found then
            Result := Void
         end
      end

   is_multipart: BOOLEAN is
      local
         f: EPX_MIME_FIELD
         ctype: STRING
      do
         f := field ("content-type")
         if f = Void then
            Result := False
         else
            create ctype.make_from_string (f.value)
            ctype.to_lower
            Result := sh.has_substring (ctype, "multipart/")
         end
      end


feature {NONE} -- parse mime header

   decode_parameters (paramline: STRING): ARRAY[EPX_MIME_PARAMETER] is
         -- parse mime header parameters
      local
         pname,
         pvalue: STRING
         start,
         stop: INTEGER
         p, q: INTEGER
         a_parameter: EPX_MIME_PARAMETER
      do

         -- its parameters
         if paramline /= Void then

            from
               create Result.make (0, -1)
               start := 1
               p := paramline.index_of ('=', start)

               -- attempt to cleanup start of paramline
               -- Opera Linux 5.0b6 emits two ;; so our paramline
               -- start with ;
               from
                  q := paramline.index_of (';', start)
               until
                  q = 0
               loop
                  if q < p then
                     start := q + 1
                     q := paramline.index_of (';', start)
                  else
                     q := 0
                  end
               end

            until
               p = 0
            loop

               -- forward to position of ; or end of line
               -- if " found, skip to next "
               -- can " itself be quoted?
               from
                  stop := p + 1
               until
                  stop >= paramline.count or else
                  paramline.item (stop) = ';'
               loop
                  if paramline.item (stop) = '"' then
                     from
                        stop := stop + 1
                     until
                        stop >= paramline.count or else
                        paramline.item (stop) = '"'
                     loop
                        stop := stop + 1
                     end
                  end
                  stop := stop + 1
               end
               if stop > paramline.count then
                  stop := paramline.count
               end

               check
                  valid_start: start < p
                  valid_p: p <= stop and paramline.item (p) = '='
                  valid_stop:
                       stop = paramline.count or
                       paramline.item (stop) = ';'
               end

               pname := sh.trim (paramline.substring (start, p-1))
               pvalue :=  sh.trim (paramline.substring (p+1, stop))

               -- remove trailing ';' character
               if pvalue.item (pvalue.count) = ';' then
                  pvalue.remove (pvalue.count)
               end

               -- remove quotation characters '"'
               if pvalue.item (1) = '"' and
                  pvalue.item (pvalue.count) = '"' then
                  pvalue.remove (pvalue.count)
                  pvalue.remove (1)
               end

               create a_parameter.make (pname, pvalue)
               Result.resize (Result.lower, Result.upper + 1)
               Result.put (a_parameter, Result.upper)

               -- prepare for next parameter, if any
               if paramline.count = stop then
                  p := 0
               else
                  start := stop + 1
                  p := paramline.index_of ('=', start)
               end
            end
         end
      end

   decode_header_field (line: STRING) is
         -- parse a single mime header field
      require
         field_available: not line.is_empty
      local
         paramline: STRING
         f: ARRAY[STRING]
         a_field: EPX_MIME_FIELD
         namevalue: STRING
         fname,
         fvalue: STRING
         parameters: ARRAY[EPX_MIME_PARAMETER]
         p: INTEGER
         found: BOOLEAN
      do
         p := line.index_of (';', 1)
         found := p /= 0
         if found then
            namevalue := line.substring (1, p-1)
            paramline := line.substring (p+1, line.count)
         else
            namevalue := line
            paramline := Void
         end

         -- entity itself
         f := sh.split_on (namevalue, ':')
         fname := sh.trim (f.item (f.lower))
         fvalue :=  sh.trim (f.item (f.upper))

         -- its parameters
         parameters := decode_parameters (paramline)
         create a_field.make (fname, fvalue, parameters)
         fields.resize (fields.lower, fields.upper + 1)
         fields.put (a_field, fields.upper)
      end

   read_mime_header is
         -- read and parse a mime header
      local
         blank_line_read,
         starts_with_blank,
         folded_line: BOOLEAN
         c: CHARACTER
         header_line: STRING
      do
         from
            header_line := sh.make_with_capacity (1024)
            blank_line_read := False
         until
            blank_line_read
         loop
            reader.read_string (1024)
            reader.chop_last_string
            blank_line_read := reader.last_string.is_empty
            if not blank_line_read then
               if reader.last_string.is_empty then
                  starts_with_blank := False
               else
                  c := reader.last_string.item (1)
                  starts_with_blank := c = ' ' or else c= '%T'
               end
               folded_line := starts_with_blank
               if folded_line then
                  -- remove leading blanks
                  reader.last_string.remove (1)
                  from
                     c := reader.last_string.item (1)
                  until
                     reader.last_string.is_empty or else
                     (c /= ' ' and c /= '%T')
                  loop
                     reader.last_string.remove (1)
                  end
                  header_line.append_character (' ')
                  header_line.append_string (reader.last_string)
               else
                  -- cached header line available?
                  if not header_line.is_empty then
                     decode_header_field (header_line)
                     sh.clear (header_line)
                  end
                  header_line.append_string (reader.last_string)
               end
            else
               if not header_line.is_empty then
                  decode_header_field (header_line)
               end
            end
         end
      end


feature {NONE} -- parse mime body

   writer_s: STRING
         -- writer buffer

   writer_file: STDC_TEMPORARY_FILE
         -- writer file

   create_singlepart_writer: EPX_WRITER is
         -- body can be written to file or stored in memory. Both
         -- cases can be handled by the correct EPX_WRITER class
      local
         f: EPX_MIME_FIELD
         parameter: EPX_MIME_PARAMETER
      do
         if save_body then
            f := field ("Content-Disposition")
            if f /= Void then
               parameter := f.parameter ("filename")
               if parameter /= Void and then not parameter.value.is_empty  then
                  create writer_file.make
                  parameter.set_file (writer_file)
                  create {EPX_FILE_WRITER} Result.make (writer_file)
               end
            end
         end

         if Result = Void then
            writer_s := sh.make_with_capacity (64)
            create {EPX_STRING_WRITER} Result.make (writer_s)
         end
      end

   rewind_singlepart_writer is
         -- close file if open
      do
         if writer_file /= Void and then writer_file.is_open then
            writer_file.seek (0)
         end
      end

   rescue_singlepart_writer is
         -- when exception has occured, clean up traces somewhat
      do
         if writer_file /= Void then
            -- close should remove this file, others have to wait
            -- until garbage collection
            writer_file.close
            writer_file := Void
         end
      end

   CRLF: string is "%N%R"

   read_singlepart_body_with_boundary (writer: EPX_WRITER) is
         -- start reading a single part. This routine has to do two
         -- things right: it has to stop when a boundary has been
         -- parsed, and the cr+lf before the boundary is not part of
         -- the file.
      require
         boundary_is_set: boundary /= Void
         has_buffer: last_line /= Void
      local
         boundary_read: BOOLEAN
         index: INTEGER
         match: BOOLEAN
         c: CHARACTER
      do
         -- this loop reads data line by line. Lines are terminated by
         -- a CRLF sequence.
         -- We have to stop when a boundary occurs somewhere in this line
         from
            boundary_read := False
            index := 1
            reader.read_character
         invariant
            1 <= index
            not boundary_read implies index <= boundary.count
            index <= last_line.count + 1
         until
            reader.eof or else
            boundary_read
         loop

            -- `c' contains the current character
            c := reader.last_character

            inspect c
            when '%R' then
               -- when a CR is encountered, we've to check for a LF to
               -- know if this is an end of line or an incidental CR
               -- (pure data)
               last_line.append_character (c)
            when '%N' then
               if last_line.count > 0 and then
                  last_line.peek_character (last_line.count - 1) = '%R' then
                  -- at this point a CRLF has been read
                  -- the next line could contain the boundary so we
                  -- don't write the CRLF, but keep it in `last_line'
                  last_line.remove_last_character
                  writer.write_pointer (last_line.ptr, last_line.count)
                  last_line.clear
                  index := 1
                  last_line.append_character ('%R')
                  last_line.append_character (c)
               else
                  -- previous character was not a CR, what has been
                  -- read up to now, including this LF is pure data
                  writer.write_pointer (last_line.ptr, last_line.count)
                  writer.write_character (c)
                  last_line.clear
                  index := 1
               end
            else
               last_line.append_character (c)
               match := boundary.item (index) = c
               if match then
                  -- match found, advance to next character
                  index := index + 1
                  boundary_read := index > boundary.count
               else
                  -- mismatch, retrack to find longest match, if any
                  if index > 2 then
                     from
                        index := index - 1
                     until
                        index = 1 or else
                        last_line.substring (last_line.count - index, last_line.count - 1).is_equal (boundary.substring (1, index))
                     loop
                        index := index - 1
                     end
                  end
               end
            end

            -- very expensive check
            check
               -- last_line.has_substring (boundary) implies boundary_read
            end

            reader.read_character
         end

         -- if boundary read, skip rest of line
         if boundary_read then
            -- read two dashes or CRLF
            -- because we might read two dashes, we append them to the
            -- last read line
            -- NOTE: we've already read one character.
            last_line.append_character (reader.last_character)
            reader.read_character
            last_line.append_character (reader.last_character)
            -- if no CRLF read, skip rest of line
            if reader.last_character /= '%R' then
               reader.read_string (0)
            end
         end

      ensure
         -- boundary_read or reader.eof
      end

   read_singlepart_body_without_boundary (writer: EPX_WRITER) is
      do
         from
            reader.read_character
         until
            reader.eof
         loop
            writer.write_character (reader.last_character)
            reader.read_character
         end
      end

   read_singlepart_body is
         -- read from input until end of file is reached or a line
         -- contains `boundary'.
      local
         writer: EPX_WRITER
      do
         writer := create_singlepart_writer
         if boundary = Void then
            read_singlepart_body_without_boundary (writer)
         else
            read_singlepart_body_with_boundary (writer)
         end

         create {EPX_MIME_BODY_OTHER} body.make (writer_s)

         rewind_singlepart_writer
      rescue
         rescue_singlepart_writer
      end

   read_mime_body (my_boundary: STRING) is
         -- input is positioned at start of mime body if single-part
         -- it is positioned upon or before the boundary if multi-part
      local
         part: EPX_MIME_PART
         part_boundary: STRING
         last_boundary: STRING
         multipart_body: EPX_MIME_BODY_MULTIPART
      do
         if is_multipart then
            part_boundary := field ("content-type").parameter ("boundary").value
            skip_to_boundary (part_boundary)
            last_boundary := part_boundary + "--"
            -- read here parts until end boundary reached
            -- add every part as body
            from
               create multipart_body.make
            until
               reader.eof or else
               last_line.has_substring (last_boundary)
            loop
               create part.make (reader)
               part.set_boundary (part_boundary)
               part.set_save_body (save_body)
               part.parse
               multipart_body.append (part)
            end
            body := multipart_body
         else
            read_singlepart_body
         end
      end

   skip_to_boundary (my_boundary: STRING) is
         -- move input cursor to line that contains `boundary'`
      do
         from
         until
            reader.eof or else
            sh.has_substring (reader.last_string, my_boundary)
         loop
            reader.read_string (128)
         end
      ensure
         input_cursor_ok: reader.eof or else
                          sh.has_substring (reader.last_string, my_boundary)
      end


feature {NONE} -- input

   reader: EPX_READER


invariant

   has_fields: fields /= Void

   has_reader: reader /= Void

end -- class EPX_MIME_PART
